'''
Computes CPHydra schedule of solvers for a given set of instances.
'''

import sys
import csv
import json
import Mistral
from Numberjack import *

# Path of auxiliary files.
FEAT_PATH = '/public/xcsp_instances/new/predictions/csp/SATzilla/features/'
PORT_FILE = '/public/xcsp_instances/new/scripts/classifiers/csp/CPHydra/portfolios_csp'
BEST_SOLVERS = '/public/xcsp_instances/new/solvers/csp/fixed/best_solvers'
INFO_TIMES = '/public/xcsp_instances/new/solvers/csp/fixed/infotimes_csp'
TIMEOUT = 1800
FEAT_SOLVER = 'gecode'
portfolios = csv.reader(open(PORT_FILE), delimiter='|', skipinitialspace = True)
with open(BEST_SOLVERS, 'r') as infile:
  bs = json.load(infile)
with open('/public/xcsp_instances/new/scripts/classifiers/csp/3S/rem_norm/feat_vector', 'r') as infile:
  feat_vector = json.load(infile)
with open(INFO_TIMES, 'r') as infile:
  infotimes = json.load(infile)

def solver_performance(inst, solv):
  it = infotimes[inst][solv]
  if it[0] == "sat" or it [0] == "unsat":
    # Here not consider features extraction time.
    return it[1]
  else:
    return TIMEOUT

def compare_instances(instance1, instance2):
  """
  compute the distance between two instances
  """
  distance = 0.0
  for i in range(0, len(feat_vector[instance1])):
    d = float(feat_vector[instance1][i]) - float(feat_vector[instance2][i])
    distance += d * d
  return distance
        
def get_similar_instances(instance, training_instances, k = 10):
  """
  return the most k similar instances in the training set
  """
  distances = []
  for i in training_instances:
    distance = compare_instances(instance, i)
    distances.append((distance,i))
  distances.sort(key = lambda x : x[0])
  return distances[:k]

def get_probabilities(instance, similar_insts_with_distance, solver):
    """
    function that has been adapted from the Numberjack implementation
    of cpHydra
    """
    epsilon = 0.0001
    W = 0
    data = {}
    for (d,c) in similar_insts_with_distance:
        if d>1+epsilon:
            t = solver_performance(c, solver)
            if t in data:
                data[t] += 1/d
            else:
                data[t] = 1/d
            W += 1/d
    runtimes = data.keys()
    runtimes.sort()
    proba = [(0,0)]
    p = 0
    for r in runtimes:
        p += data[r]
        proba.append((r, p/W))
    return proba
    
def schedule(instance, cases, solvers):
    """
    function that has been adapted from the Numberjack implementation
    of cpHydra
    """

    def weight(case):
        d = case[0]
        if d >= 1000: return 1
        elif d <= 0.01: return 100000
        else: return int(1000/d)

    model = Model()

    allocated_times = VarArray(len(solvers), 0, TIMEOUT)
    prob_solved = VarArray(len(cases))

    model.add(Sum(allocated_times) == int(TIMEOUT))
    
    for prob, case in zip(prob_solved, cases):
        expected_times = []
        for solver in solvers:
            expected_times = [solver_performance(case[1], solver) for solver in solvers]
            model.add(prob == (Sum([stime >= int(etime+1) for stime, etime in
                                zip(allocated_times, expected_times)]) >= 1))
    
    model.add(Maximise(Sum(prob_solved, [weight(case) for case in cases])))
    
    solver = Mistral.Solver(model, allocated_times)
    
    solver.setHeuristic('MinDomain','RandomSplit',3)
    
    solver.solveAndRestart()

    return dict(zip(solvers, map(value, allocated_times)))

def compute_solving_times(solvers, instance, train_set):
  """
  given the list of solvers, an instance, and the list of the 
  training instances it computes the solving time of cpHydra
  """
  # Features extraction time.
  time = float(bs[instance]['feat'])
  #print "computing map for instance " + instance
  similar_instances = get_similar_instances(instance, train_set)
  solvers_map = schedule(instance, similar_instances, solvers)
  # Simulate portfolio execution and print the results.
  for solver in solvers:
    perf_time = solver_performance(instance, solver)
    # Not reconvert in FlatZinc if already done.
    if solver == FEAT_SOLVER:
      perf_time -= float(bs[instance]['mzn2fzn'])
    if solvers_map[solver] >= perf_time:
      time += perf_time
      if time < TIMEOUT:
        print instance + "|1|" + str(time) + "|" + str(solvers_map)
      else:
        print instance + "|0|" + str(TIMEOUT) + "|" + str(solvers_map)
      return
    else:
      time += solvers_map[solver]
  print instance + "|0|" + str(TIMEOUT) + "|" + str(solvers_map)

def compute_predictions(solvers, rep, fold, f, t):
  """
  compute the desired statistics of the cpHydra simulation
  """
  # compute the instances of the train/test set.
  train_file = FEAT_PATH + 'train_rep' + rep + '_fold' + fold + '.features'
  test_file = FEAT_PATH + 'test_rep' + rep + '_fold' + fold + '.features'
  train_reader = csv.reader(open(train_file, "rb"), delimiter=' ', skipinitialspace = True)
  test_reader = csv.reader(open(test_file, "rb"), delimiter=' ', skipinitialspace = True)
  train_set = []
  test_set = []
  for row in train_reader:
    train_set.append(row[0])
  for row in test_reader:
    test_set.append(row[0])
  for i in range(min(len(test_set) - 1, t) - f + 1):
    compute_solving_times(solvers, test_set[f + i], train_set)
        
def main(args):
  n_solver = args[0]
  for row in portfolios:
    if row[0] == n_solver:
      portfolio = eval(row[1])
      break
  rep  = args[1]
  fold = args[2]
  f    = int(args[3])
  t    = int(args[4])
  compute_predictions(portfolio, rep, fold, f, t)

if __name__ == '__main__':
    main(sys.argv[1:])
