'''
WRITE HELP
'''

import sys
import os
import getopt
from subprocess import Popen, PIPE, call
from math import isnan, isinf, sqrt
import json
import signal
from combinations import *

MZN2FEAT = '/public/xcsp_instances/sunny-csp/mzn2feat/mzn2feat'
BACKUP_SOLVER = 'minisatid'
FEAT_SOLVER = 'gecode'
TIMEOUT = 1800.0
# Rank of the solvers of the portfolio sorted by decreasing PSI.
SOLVERS_RANK = [
  'minisatid', 'fzn2smt', 'g12cpx', 'g12lazyfd', 'izplus', 'ortools', 
  'gecode', 'jacop', 'bprolog', 'g12fd', 'mistral', 'g12mip'
]
CHILD_PROC = None
ACTUAL_CMD = None

def handler(signum = None, frame = None):
  """
  Handles termination signals.
  """
  print '*** Signal handler called with signal',signum
  if CHILD_PROC:
    CHILD_PROC.terminate()
  print '=====UNKNOWN====='
  for c in ACTUAL_CMD:
    cmd = 'pkill -9 ' + c.split()[0]
    call(cmd.split())
  sys.exit(1)
  
for sig in [signal.SIGTERM, signal.SIGINT, signal.SIGHUP, signal.SIGQUIT]:
  signal.signal(sig, handler)

def extract_features(mzn, dzn, fzn):
  """
  Extract the features from a MiniZinc model by exploiting the mzn2feat
  features extractor.
  """
  cmd = ['bash', MZN2FEAT, mzn, dzn, fzn]
  global CHILD_PROC, ACTUAL_CMD
  CHILD_PROC = Popen(
    cmd, stdout = PIPE, stderr = PIPE
   )
  ACTUAL_CMD = ['mzn2fzn', 'fzn2feat']
  (out, err) = CHILD_PROC.communicate()
  ret = CHILD_PROC.returncode
  if ret not in [0, 3, 4, 5, 8]:
    sys.stderr.write(str(err))
    print '*** Extraction failed! Trying other solvers...'
    if ret in [6, 7]:
      solvers = [FEAT_SOLVER]
      solvers.extend([s for s in SOLVERS_RANK if s != 'gecode'])
    else:
      solvers = SOLVERS_RANK
    for solver in solvers:
      print '*** Executing',solver,'...'
      exe_solver(solver, TIMEOUT, mzn, dzn, fzn)
    print '=====UNKNOWN====='
    print '*** Instance Not Solved by any solver'
    sys.exit(1)
  elif ret == 3:
    print out
    print '*** Instance Solved by',FEAT_SOLVER
    sys.exit(0)
  features = out.split("|")
  feat_vector = [float(features[i]) for i in range(1, len(features))]
  return feat_vector

def normalize_features(feat_vector, lims):
  """
  Given a features vector, it returns a normalized one in which constant 
  features are removed and features values are scaled in [-1, 1], by 
  exploiting the information already computed in lims.
  """
  norm_vector = []
  for i in range(0, len(feat_vector)):
    if lims[i]:
      val = float(feat_vector[i])
      min_val = float(lims[i][0])
      max_val = float(lims[i][1])
      if isnan(val) or isinf(val):
        norm_val = -1
      elif val <= min_val:
        norm_val = -1
      elif val >= max_val:
        norm_val = 1
      else:
        x = (val - min_val) / (max_val - min_val)
        norm_val = 2 * x - 1
        assert (-1 <= norm_val and norm_val <= 1)
      norm_vector.append(norm_val)
  return norm_vector

def compare_instances(fv1, fv2):
  """
  Compute the Euclidean distance between two instances.
  """
  distance = 0.0
  for i in range(0, len(fv1)):
    d = fv1[i] - fv2[i]
    distance += d * d
  return sqrt(distance)

def find_neighbours(fv, feat_vector, k):
  """
  Return the most k similar instances in the training set.
  """
  distances = []
  for (inst, vect) in feat_vector.items():
    if inst != 'LIMS':
      distance = compare_instances(fv, vect)
      distances.append((distance, inst))
  distances.sort(key = lambda x : x[0])
  return [i for (d, i) in distances[:k]]

def solver_performance(instance, solver, runtimes):
  """
  Return the runtime of a solver on a given instances
  """
  it = runtimes[instance][solver]
  if it[0] == "sat" or it [0] == "unsat":
    # Here not consider the features extraction time.
    return it[1]
  else:
    return TIMEOUT

def compute_schedule(neighbours, runtimes, k):
  """
  Given the neighborhood of a given instance, return the corresponding 
  SUNNY schedule.
  """
  solved = {}
  times  = {}
  portfolio = SOLVERS_RANK
  n_solv = len(portfolio)
  for solver in portfolio:
    solved[solver] = set([])
    times[solver]  = 0.0
  for inst in neighbours:
    for solver in portfolio:
      time = solver_performance(inst, solver, runtimes)
      if time < TIMEOUT:
        solved[solver].add(inst)
      times[solver] += time
  max_solved = 0
  min_time = 1000000
  best_portfolio = []
  # Select the best sub-portfolio.
  for i in range(0, n_solv + 1):
    for j in range(0, binom(n_solv, i)):
      solved_instances = set([])
      solving_time = 0
      sub_portfolio = get_subset(j, i, portfolio)
      for solver in sub_portfolio:
        solved_instances.update(solved[solver])
        solving_time += times[solver]
      num_solved = len(solved_instances)
      if num_solved >  max_solved or \
        (num_solved == max_solved and solving_time < min_time):
          min_time = solving_time
          max_solved = num_solved
          best_portfolio = sub_portfolio
  # n is the number of instances solved by each solver plus the instances 
  # that no solver can solver.
  n = sum([len(solved[s]) for s in best_portfolio]) + (k - max_solved)
  schedule = {}
  # Compute the schedule and sort it by number of solved instances.
  for solver in best_portfolio:
    ns = len(solved[solver])
    schedule[solver] = TIMEOUT / n * ns
  tot_time = sum(schedule.values())
  # Allocate to the backup solver the (eventual) remaining time.
  if round(tot_time) < TIMEOUT:
    if BACKUP_SOLVER in schedule.keys():
      schedule[BACKUP_SOLVER] += TIMEOUT - tot_time
    else:
      schedule[BACKUP_SOLVER]  = TIMEOUT - tot_time
  sorted_schedule = sorted(
    schedule.items(), 
    key = lambda x: len(solved[x[0]]), 
    reverse = True
  )
  assert(round(sum(t for (s,t) in sorted_schedule)) <= TIMEOUT)
  return sorted_schedule

def exe_schedule(mzn, dzn, fzn, schedule):
  """
  Executes the SUNNY schedule on a MiniZinc model, possibly exploiting .dzn 
  data and the Gecode FlatZinc model computed during features extraction.
  """
  rem_time = 0.0
  timeout  = 0.0
  for (solver, timeout) in schedule:
    timeout += rem_time
    print '*** Executing',solver,'for',str(timeout),'seconds...'
    rem_time = exe_solver(solver, timeout, mzn, dzn, fzn)
  print '*** Instance not solved by scheduled solvers! Trying others...'
  sched_solvers = [s for (s, t) in schedule]
  other_solvers = [s for s in SOLVERS_RANK if s not in sched_solvers]
  for solver in other_solvers:
    print '*** Executing',solver,'...'
    exe_solver(solver, TIMEOUT, mzn, dzn, fzn)
  print '=====UNKNOWN====='
  print '*** Instance Not Solved by any solver'
  sys.exit(1)

def exe_solver(solver, timeout, mzn, dzn, fzn):
  """
  Executes a single solver for a given timeout on a given MiniZinc model.
  """
  global CHILD_PROC, ACTUAL_CMD
  if solver != FEAT_SOLVER:
    # FlatZinc conversion.
    if dzn == 'NODATA':
      dzn = ''
    cmd = 'time -p timeout ' + str(timeout) + ' '
    mzn2fzn = cmd_mzn2fzn(solver, mzn, dzn, fzn)
    if not mzn2fzn:
      sys.stderr.write("*** Error: Undefined Solver!\n")
      return timeout
    cmd += ' ' + mzn2fzn
    mzn_dir = os.path.dirname(mzn)
    CHILD_PROC = Popen(
      cmd.split(), stdout = PIPE, stderr = PIPE, cwd = mzn_dir
    )
    ACTUAL_CMD = ['mzn2fzn']
    (out, err) = CHILD_PROC.communicate()
    times = err.split()[-6:len(err)]
    time = float(times[1])
    timeout -= time
    if CHILD_PROC.returncode != 0:
      sys.stderr.write(str(err))
      sys.stderr.write("*** Error: MiniZinc model not converted!\n")
      return timeout
  # Executes the FlatZinc model.
  cmd = 'time -p timeout ' + str(timeout) + ' '
  flatzinc = cmd_flatzinc(solver, fzn)
  if not flatzinc:
    sys.stderr.write("*** Error: Undefined Solver!\n")
    return timeout
  cmd += flatzinc
  CHILD_PROC = Popen(
    cmd.split(), stdout = PIPE, stderr = PIPE
  )
  ACTUAL_CMD = [flatzinc]
  if solver == 'fzn2smt':
    ACTUAL_CMD.append('yices')
  elif solver == 'ortools':
    ACTUAL_CMD.append('fz')
  (out, err) = CHILD_PROC.communicate()
  if '----------' in out or '=====UNSATISFIABLE=====' in out:
    print out
    print '*** Instance Solved by',solver
    sys.exit(0)
  else:
    sys.stderr.write(str(err))
    print '*** Instance Not Solved by',solver
    times = err.split()[-6:len(err)]
    time = float(times[1])
    if timeout - time < 0:
      timeout -= time
    else:
      timeout = 0.0
    return timeout
 
def cmd_mzn2fzn(solver, mzn, dzn, fzn):
  """
  Returns the command to be run for converting a MiniZinc model into FlatZinc.
  """
  cmd = ''
  if solver == "g12cpx":
    cmd += 'mzn2fzn -G g12_cpx ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "g12fd":
    cmd += 'mzn2fzn -G g12_fd ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "g12lazyfd":
    cmd += 'mzn2fzn -G g12_lazyfd ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "g12mip":
    cmd += 'mzn2fzn -G linear ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "minisatid":
    cmd += 'mzn2fzn ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "fzn2smt":
    cmd += 'mzn2fzn ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "bprolog":
    mznlib = os.environ['HOME'] + '/new/solvers/BProlog/mznlib/'
    cmd += 'mzn2fzn -I ' + mznlib + ' ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "izplus":
    mznlib = os.environ['HOME'] + '/new/solvers/izplus20121024/tmplib'
    cmd += 'mzn2fzn -I ' + mznlib + ' ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "ortools":
    mznlib = os.environ['HOME'] \
           + '/new/solvers/or-tools-read-only/src/flatzinc/mznlib'
    cmd += 'mzn2fzn -I ' + mznlib + ' ' + mzn + ' ' + dzn + ' -o ' + fzn
  elif solver == "mistral":
    mznlib = os.environ['HOME'] + '/new/solvers/Mistral-2.0/fz/mznlib'
    cmd += 'mzn2fzn -I ' + mznlib + ' ' + mzn + ' ' + dzn + ' -o ' + fzn
  return cmd
    
def cmd_flatzinc(solver, fzn):
  """
  Returns the command to be run for executing a FlatZinc model.
  """
  cmd = ''
  if solver == 'gecode':
    cmd += 'fz ' + fzn
  elif solver == "g12cpx":
    cmd += 'fzn_cpx ' + fzn
  elif solver == "g12fd":
    cmd += 'flatzinc ' + fzn
  elif solver == "g12lazyfd":
    cmd += 'flatzinc -b lazy ' + fzn
  elif solver == "g12mip":
    cmd += 'flatzinc -b mip ' + fzn
  elif solver == "minisatid":
    cmd += 'minisatid -f=fz ' + fzn
  elif solver == "fzn2smt":
    cmd += 'fzn_fzn2smt -f ' + fzn
  elif solver == "bprolog":
    fz = os.environ['HOME'] + '/new/solvers/BProlog/flatzinc.out'
    cmd += 'bp ' + fz + ' -g \"fzn_main,halt\" ' + fzn
  elif solver == "izplus":
    cmd += 'fz_izplus -f ' + fzn
  elif solver == "ortools":
    fz = os.environ['HOME'] + '/new/solvers/or-tools-read-only/bin/fz '
    cmd += fz + fzn
  elif solver == "mistral":
    cmd += 'fzn-mistral ' + fzn
  return cmd
 
def main(args):
  n_args = len(args)
  k = 10
  if n_args < 9:
    print "Wrong number of arguments! For help use --help"
    sys.exit(1)
  # MiniZinc model.
  mzn = args[0]
  dzn = 'NODATA'
  try:
    opts, args = getopt.getopt(args[1 : n_args], "hd:f:r:k:z:", ["help"])
  except getopt.error, msg:
    print msg
    print "For help use --help"
    sys.exit(2)
    # process options
  for o, a in opts:
    if o in ("-h", "--help"):
      print 'Documentation (TODO!).' #__doc__
      sys.exit(0)
    elif o in ("-k"):
      # k parameter.
      k = int(a)
    elif o in ("-f"):
      # Features file.
      with open(a, 'r') as infile:
        feat_vector = json.load(infile)
    elif o in ("-r"):
      # Runtimes file.
      with open(a, 'r') as infile:
        runtimes = json.load(infile)
    elif o in ("-d"):
      # Data model.
      dzn = a
    elif o in ("-z"):
      # FlatZinc model.
      fzn = a
  
  print '*** Extracting features...'
  fv = extract_features(mzn, dzn, fzn)
  print '*** Normalizing features...'
  fv = normalize_features(fv, feat_vector['LIMS'])
  print '*** Finding the ' + str(k) + ' nearest neighbours...'
  neighbours = find_neighbours(fv, feat_vector, k)
  print '*** Computing sub-portfolios and solvers schedule...'
  schedule = compute_schedule(neighbours, runtimes, k)
  print '*** Solvers schedule:',schedule
  print '*** Executing schedule...'
  exe_schedule(mzn, dzn, fzn, schedule)

if __name__ == '__main__':
    main(sys.argv[1:])
